import pyadchpp as adchpp
import SettingManager
import UserManager
import general
import Hub
import autil
import re, shlex
import threading
import time
import access_ADCCMDMgr
#import json

commands = {}

#Used for ordering in help menu
command_noCategory = {}
command_categories = {}

curr_cmdchar = "!"
curr_pubcmdchar = ""
cmdRegex = re.compile("^[!] ?[^\s]\w* ?(.*)")
pubcmdRegex = re.compile("^[+] ?[^\s]\w* ?(.*)")
cmd = None
ucmd_cmd = None

context_hub = 0x1
context_userlist = 0x2
context_search = 0x4
context_filelist = 0x8

cmdLock = threading.Lock()

class Command(object):
    def __init__(self, key = None, alias = [], command = None, helpTxt = "", helpTxtLong = "", minLevel = 0, category = "", user_command = []):        
        self.key = key
        self.alias = alias
        self.command = command
        self.helpTxt = helpTxt
        self.helpTxtLong = helpTxtLong
        self.minLevel = minLevel
        self.category = category
        if type(user_command) is list:
            self.user_command = user_command
        elif type(user_command) is UserCommand:
            self.user_command = [user_command]
        else:
            self.user_command = []

    def release(self):
        remove(self.key)

class UserCommand(object):
    def __init__(self, contexts = 0, command = "", path = "", hasPublic = False):        
        self.contexts = contexts
        self.command = command
        self.path = path
        self.hasPublic = hasPublic

#options = {alias, command, helpTxt, helpTxtLong, minLevel, user_command} #user_command = {contexts, command, path, hasPublic}

def initModule():
    SettingManager.add(SettingManager.Setting(key="cmd_chr", change=compileCmdRegex,helpTxt='The character which starts off all commands, i.e. !command or +command',value='!',validate=validateCmdChr))
    SettingManager.add(SettingManager.Setting(key="pubcmd_chr", change=compileCmdRegex_pub,helpTxt='The character which acts like the normal command char, but some scripts may send their messages publicly instead of privately, i.e. !command or +command. Enter nothing to disable', value='', validate=validatePubCmdChr))
    #compileCmdRegex()
    global cmd, ucmd_cmd
    cmd = autil.handleCommand(onMSG, adchpp.AdcCommand.CMD_MSG)
    ucmd_cmd = autil.handleState(send_ucmds,adchpp.Entity.STATE_NORMAL)
    add(Command(key="help",command=cmdHelp, helpTxt="[command] - list all the available commands or display detailed information about one specific command",
                 user_command=UserCommand(contexts=context_hub + context_userlist,command='help', path='help')))

def getMinLevel(cmd):
    if cmd.minLevel:
        if type(cmd.minLevel) is int:
            return cmd.minLevel
        elif type(cmd.minLevel) is SettingManager.Setting:
            if type(cmd.minLevel.value) is int:
                return cmd.minLevel.value

    return 0

def cmdHelp(ent, cmd, params, public):
    def command_help(command):
        cmdhelpStr = curr_cmdchar + command.key
        if command.helpTxt:
            cmdhelpStr += " " + command.helpTxt
        if command.alias and len(command.alias) > 0:
            cmdhelpStr += " (aliases: "
            for k in command.alias:
                cmdhelpStr += curr_cmdchar + k + ", "
            cmdhelpStr = cmdhelpStr[:len(cmdhelpStr)-2] + ")"
                
        return cmdhelpStr
    
    if len(params) > 0:
        key = None
        
        cmd = str(params[0]).lower()
        
        for k in commands:
            if k == cmd or (commands[k].alias and cmd in commands[k].alias):
                userLevel = UserManager.getLevel(ent)
                #if 'minLevel' not in commands[k] or ('minLevel' in commands[k] and userlevel >= int(commands[k].minLevel)):
                if userLevel >= getMinLevel(commands[k]):
                      key = k
                      break

        if not key:
            autil.hub_reply(ent, "The command " + cmd + " cannot be found")
            return

        replyText = "\n" + command_help(commands[key])
        if commands[key].helpTxtLong:
            replyText = replyText + "\n\n"
            if callable(commands[key].helpTxtLong):
                replyText += str(commands[key].helpTxtLong())
            else:
                replyText += str(commands[key].helpTxtLong)
        autil.hub_reply(ent, replyText)
    else:#show help menu
        replyText = "Available commands:\n"
        cmdItems = command_noCategory.items()
        cmdItems.sort()
        for k in cmdItems:
            command = k[1]
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(command):
                replyText += command_help(command) + "\n"

        cmdItems = command_categories.items()
        cmdItems.sort()
        for cat in cmdItems:
            catItems = cat[1].items()
            catItems.sort()
            atLeastOneCommand = False
            almostReplyText = "\r" + "[" + cat[0] + "]\n"
            for c in catItems:
                command = c[1]
                userLevel = UserManager.getLevel(ent)
                if userLevel >= getMinLevel(command):
                    atLeastOneCommand = True
                    almostReplyText += command_help(command) + "\n"
            if atLeastOneCommand:
                replyText += almostReplyText
        """cmdItems = commands.items()
        cmdItems.sort()
        for k in cmdItems:
            key = k[0]
            value = k[1]
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(value):
                replyText += command_help(key, value) + "\n"""

        if curr_pubcmdchar:
            replyText += "\n\nSome commands will send a message to everyone in the hub when you use the starting command character: " + curr_pubcmdchar
        autil.hub_reply(ent, replyText)
    return(False)

def onMSG(ent, cmd, ok):
    #to = cmd.getTo()
    #if to == adchpp.AdcCommand.INVALID_SID or to == adchpp.AdcCommand.HUB_SID:
    if cmd.getParam(0) == "+stats": #because whoever wrote bloom is stupid and likes +
        return True
    public = curr_pubcmdchar != "" and re.match(pubcmdRegex, cmd.getParam(0)) != None
    #print(cmd.getParam(0))
    if public or re.match(cmdRegex, cmd.getParam(0)):
        #print(re.split(r'\s+', cmd.getParam(0)[1:],1))
        #params = shlex.split(cmd.getParam(0)[1:])
        #print(cmd.getParam(0))
        params = strSplitFirstWhite(cmd.getParam(0)[1:])
        command = params.pop(0).lower()

        foundCommand = None
        if command in commands:
            foundCommand = commands[command]
        else:
            for k in commands:
                if k == command or (commands[k].alias and command in commands[k].alias):
                    foundCommand = commands[k]
        
        if foundCommand:
            access_ADCCMDMgr.addHubCmdStats(foundCommand.key)
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(foundCommand):
                #cmd.setPriority(adchpp.AdcCommand.PRIORITY_IGNORE)
                #foundCommand.command(ent, cmd, params, public)
                commandRunner(foundCommand.command, ent, cmd, params, public)
                return False
                #return foundCommand.command(ent, cmd, params, public)
        #cmd.setPriority(adchpp.AdcCommand.PRIORITY_IGNORE)
        autil.hub_reply(ent, "Command not recognized")
        return False
    return True

def strSplitFirstWhite(string):
    return(re.split(r'\s', string,1))

def validateCmdChr(new):
    if len(new) != 1:
        return("The length of the cmdchar must be 1")

    if not re.match(r'^\W',new):
        return("The cmdchar must not be an alphanumeric character")

def validatePubCmdChr(new):
    if len(new) > 1:
        return("The length of the pubcmdchar must be 1 or 0")

    if len(new) != 0 and not re.match(r'^\W',new):
        return("The cmdchar must not be an alphanumeric character")

def compileCmdRegex(old_cmdchar, newValue):
    global cmdRegex, curr_cmdchar

    old_cmdchar = curr_cmdchar
    curr_cmdchar = newValue

    if curr_cmdchar:
        cmdRegex = re.compile('^[' + curr_cmdchar + "] ?[^\s]\w* ?(.*)")

    #remove all ucmds and add again
    for key in commands:
        if commands[key].user_command:
            send_rm_ucmd_to_all(commands[key], old_cmdchar, None)
            send_ucmd_to_all(commands[key], curr_cmdchar, None)

def compileCmdRegex_pub(old_pubcmdchar, newValue):
    global pubcmdRegex, curr_pubcmdchar

    curr_cmdchar = None
    curr_pubcmdchar = newValue

    if 'pubcmd_chr' in SettingManager.settings:
        curr_pubcmdchar = str(SettingManager.settings['pubcmd_chr'].value)

    if curr_pubcmdchar:
        pubcmdRegex = re.compile('^[' + curr_pubcmdchar + "] ?[^\s]\w* ?(.*)")

    #remove all ucmds and add again
    for key in commands:
        if commands[key].user_command:
            send_rm_ucmd_to_all(commands[key], None, old_pubcmdchar)
            send_ucmd_to_all(commands[key], None, curr_pubcmdchar)

def createLine(desc):
    return("%[line:" + desc + "]")

def send_ucmds(ent, oldState):
    if ent.hasSupport(adchpp.AdcCommand.toFourCC("UCMD")) or ent.hasSupport(adchpp.AdcCommand.toFourCC("UCM0")):
        for key in commands:
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(commands[key]):
                if commands[key].user_command:
                    for ucmd in commands[key].user_command:
                        send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
                        return_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_HUB, ent.getSID())

                        return_ucmd.addParam(curr_cmdchar + ucmd.command)
                        send_ucmd.addParam(ucmd.path)
                        send_ucmd.addParam("TT", return_ucmd.toString())
                        send_ucmd.addParam("CT", str(ucmd.contexts))
                        ent.send(send_ucmd);

                        if curr_pubcmdchar and ucmd.hasPublic:
                            send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
                            return_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_HUB, ent.getSID())

                            return_ucmd.addParam(curr_pubcmdchar + ucmd.command)
                            send_ucmd.addParam(ucmd.path+"_public")
                            send_ucmd.addParam("TT", return_ucmd.toString())
                            send_ucmd.addParam("CT", str(ucmd.contexts))
                            ent.send(send_ucmd);
    return True

def send_ucmd_to_all(cmd, cmdchar, pubcmdchar):
    ucmds = cmd.user_command
    ent_list = Hub.ClientMgr.getEntities();
    for ent in ent_list:
        if ent.hasSupport(adchpp.AdcCommand.toFourCC("UCMD")) or ent.hasSupport(adchpp.AdcCommand.toFourCC("UCM0")):
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(cmd):
                for ucmd in ucmds:
                    #private
                    if cmdchar:
                        send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
                        return_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_HUB, ent.getSID())

                        return_ucmd.addParam(cmdchar + ucmd.command)
                        send_ucmd.addParam(ucmd.path)
                        send_ucmd.addParam("TT", return_ucmd.toString())
                        send_ucmd.addParam("CT", str(ucmd.contexts))
                        ent.send(send_ucmd);

                    #public
                    if pubcmdchar and ucmd.hasPublic:
                        send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
                        return_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_HUB, ent.getSID())

                        return_ucmd.addParam(pubcmdchar + ucmd.command)
                        send_ucmd.addParam(ucmd.path+"_public")
                        send_ucmd.addParam("TT", return_ucmd.toString())
                        send_ucmd.addParam("CT", str(ucmd.contexts))
                        ent.send(send_ucmd);

def send_rm_ucmd_to_all(cmd, cmdchar, pubcmdchar):
    ucmds = cmd.user_command
    ent_list = Hub.ClientMgr.getEntities();
    for ent in ent_list:
        if ent.hasSupport(adchpp.AdcCommand.toFourCC("UCMD")) or ent.hasSupport(adchpp.AdcCommand.toFourCC("UCM0")):
            userLevel = UserManager.getLevel(ent)
            if userLevel >= getMinLevel(cmd):
                for ucmd in ucmds:
                    if cmdchar:
                        send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)

                        send_ucmd.addParam(ucmd.path)
                        send_ucmd.addParam("RM", "1")
                        send_ucmd.addParam("CT", str(ucmd.contexts))
                        ent.send(send_ucmd);

                    if pubcmdchar and ucmd.hasPublic:
                        send_ucmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_CMD, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
                        
                        send_ucmd.addParam(ucmd.path+"_public")
                        send_ucmd.addParam("RM", "1")
                        send_ucmd.addParam("CT", str(ucmd.contexts))
                        ent.send(send_ucmd);

def add(command): #options = {alias, command, helpTxt, helpTxtLong, minLevel, category, user_command} #user_command = {contexts, command, path, hasPublic}
    if command.command != None:
        commands[command.key.lower()] = command
        if command.alias:
            command.alias = [x.lower() for x in command.alias]
        if command.category:
            if not command.category in command_categories:
                command_categories[command.category] = {}
            if not command.key.lower() in command_categories[command.category]:
                command_categories[command.category][command.key.lower()] = command
        else:
            if not command in command_noCategory:
                command_noCategory[command.key.lower()] = command
        if command.user_command != None and len(command.user_command) > 0:
            send_ucmd_to_all(command, curr_cmdchar, curr_pubcmdchar)

def remove(key):
    if key in commands:
        if commands[key].user_command:
            send_rm_ucmd_to_all(commands[key], curr_cmdchar, curr_pubcmdchar)
        if commands[key].category:
            if commands[key].category in command_categories and commands[key].key.lower() in command_categories[commands[key].category]:
                del(command_categories[commands[key].category][commands[key].key.lower()])
                if not command_categories[commands[key].category]:
                    del(command_categories[commands[key].category])
        else:
            if command in command_noCategory:
                command_noCategory.remove(command)
        del(commands[key])

class commandRunner(threading.Thread):
    def __init__(self, func, ent, cmd, params, public):
        threading.Thread.__init__(self)
        self.func = func
        self.ent = ent
        
        self.cmd = adchpp.AdcCommand() #cmd gets destroyed after caller function returns value
        self.cmd.parse(cmd.toString()) #cmd probably won't exist when thread runs, copy it
        
        self.params = params
        self.public = public
        self.daemon = True
        self.start()
        return
    
    #def __del__(self):
        #return
    
    def run(self):
        try:
            autil.threadData.origUser = UserManager.findUser(self.ent.getField('NI'), self.ent.getCID().toBase32())
            autil.threadData.replyFrom = self.cmd.getTo()
            self.func(self.ent, self.cmd, self.params, self.public)
        except Exception as err:
            autil.log(__name__, str(err))
        return
