import pyadchpp as adchpp
import Hub
import UserManager
import SettingManager
import CommandManager
import ScriptManager
import autil
import re, time

utilReasons = {
    adchpp.Util.REASON_BAD_STATE: "BAD_STATE",
    adchpp.Util.REASON_CID_CHANGE: "CID_CHANGE",
    adchpp.Util.REASON_CID_TAKEN: "CID_TAKEN",
    adchpp.Util.REASON_FLOODING: "FLOODING",
    adchpp.Util.REASON_HUB_FULL: "HUB_FULL",
    adchpp.Util.REASON_INVALID_COMMAND_TYPE: "INVALID_COMMAND_TYPE",
    adchpp.Util.REASON_INVALID_IP: "INVALID_IP",
    adchpp.Util.REASON_INVALID_SID: "INVALID_SID",
    adchpp.Util.REASON_LOGIN_TIMEOUT: "LOGIN_TIMEOUT",
    adchpp.Util.REASON_MAX_COMMAND_SIZE: "MAX_COMMAND_SIZE",
    adchpp.Util.REASON_NICK_INVALID: "NICK_INVALID",
    adchpp.Util.REASON_NICK_TAKEN: "NICK_TAKEN",
    adchpp.Util.REASON_NO_BASE_SUPPORT: "NO_BASE_SUPPORT",
    adchpp.Util.REASON_NO_TIGR_SUPPORT: "NO_TIGR_SUPPORT",
    adchpp.Util.REASON_PID_MISSING: "PID_MISSING",
    adchpp.Util.REASON_PID_CID_LENGTH: "PID_CID_LENGTH",
    adchpp.Util.REASON_PID_CID_MISMATCH: "PID_CID_MISMATCH",
    adchpp.Util.REASON_PID_WITHOUT_CID: "PID_WITHOUT_CID",
    adchpp.Util.REASON_PLUGIN: "PLUGIN:",
    adchpp.Util.REASON_WRITE_OVERFLOW: "WRITE_OVERFLOW",
    adchpp.Util.REASON_NO_BANDWIDTH: "NO_BANDWIDTH",
    adchpp.Util.REASON_INVALID_DESCRIPTION: "INVALID_DESCRIPTION",
    adchpp.Util.REASON_WRITE_TIMEOUT: "WRITE_TIMEOUT",
    adchpp.Util.REASON_SOCKET_ERROR: "SOCKET_ERROR",
    adchpp.Util.REASON_LAST: "REASON_LAST"}


context_hub = "[H]"
context_bcast = "[BF]"
context_direct = "[DE]"
context_send = "[BDEFH]"
context_hubdirect = "[HDE]"

dispatch_stats = False

command_contexts = {
    adchpp.AdcCommand.CMD_STA : context_hubdirect,
    adchpp.AdcCommand.CMD_SUP : context_hub,
    adchpp.AdcCommand.CMD_SID : context_hub,
    adchpp.AdcCommand.CMD_INF : context_bcast,
    adchpp.AdcCommand.CMD_MSG : context_send,
    adchpp.AdcCommand.CMD_SCH : context_send,
    adchpp.AdcCommand.CMD_RES : context_direct,
    adchpp.AdcCommand.CMD_CTM : context_direct,
    adchpp.AdcCommand.CMD_RCM : context_direct,
    adchpp.AdcCommand.CMD_NAT : context_direct,
    adchpp.AdcCommand.CMD_RNT : context_direct,
    adchpp.AdcCommand.CMD_GPA : context_hub,
    adchpp.AdcCommand.CMD_PAS : context_hub,
    adchpp.AdcCommand.CMD_QUI : context_hub,
    adchpp.AdcCommand.CMD_GET : context_hub,
    adchpp.AdcCommand.CMD_GFI : context_hub,
    adchpp.AdcCommand.CMD_SND : context_hub}

stats = {}
HubCommandStats = {}
restrCMDs = {}
DCReasons = {}
socketErrors = {}

recieveCMD = None
disconnectCMD = None
    
def initModule():
    global stats, HubCommandStats, restrCMDs, DCReasons, socketErrors, recieveCMD, disconnectCMD

    stats = {}
    HubCommandStats = {}
    restrCMDs = {}
    DCReasons = {}
    socketErrors = {}
        
    recieveCMD = autil.handleCommand(onRecieve)
    disconnectCMD = Hub.ClientMgr.signalDisconnected().connect(onDisconnect)
        
    SettingManager.add(SettingManager.Setting(key="minchat_level", helpTxt="minimum level to chat", value=0))
    restrCMDs[adchpp.AdcCommand.CMD_MSG] = {'level':SettingManager.settings['minchat_level'], 'msg':'chat'}

    SettingManager.add(SettingManager.Setting(key="mindownload_level", helpTxt="minimum level to download", value=0))
    restrCMDs[adchpp.AdcCommand.CMD_CTM] = {'level':SettingManager.settings['mindownload_level'], 'msg':'download'}
    restrCMDs[adchpp.AdcCommand.CMD_RCM] = {'level':SettingManager.settings['mindownload_level'], 'msg':'download'}

    SettingManager.add(SettingManager.Setting(key="minsearch_level", helpTxt="minimum level to search", value=0))
    restrCMDs[adchpp.AdcCommand.CMD_SCH] = {'level':SettingManager.settings['minsearch_level'], 'msg':'search'}

    CommandManager.add(CommandManager.Command(key='info', helpTxt="[nick or CID or IP] - information about a user, or about the hub if no parameter given", command=infoCommand))

def onDisconnect(ent, reason, info):
    if reason == adchpp.Util.REASON_SOCKET_ERROR:
        if info in socketErrors:
            socketErrors[info] += 1
        else:
            socketErrors[info] = 1
                
    if reason in DCReasons:
        DCReasons[reason] += 1
    else:
        DCReasons[reason] = 1

    if ent.getState() == adchpp.Entity.STATE_NORMAL:
        user = UserManager.findUser(ent.getField('NI'), ent.getCID().toBase32())
        if user:
            user.lastofftime = time.time()
            UserManager.userChangesMade()

def infoCommand(ent, cmd, params, public):
    global dispatch_stats

    if dispatch_stats:
        return

    strInfo = ""
    #print(params)
    #params = params[0]
    if len(params) > 0:
        params = params[0]
        user = Hub.ClientMgr.findByNick(params) # by nick
        if not user:
            user = Hub.ClientMgr.findByCID(adchpp.CID(params)) # by CID

        if user:
            def field_function(field, text):
                if user.hasField(field):
                    return text + ": " + user.getField(field) + "\n"
                return ""

            strInfo = "\n"
            strInfo += field_function("NI", "Nick")
            strInfo += field_function("ID", "CID")
            strInfo += "IP: "
            user_c = user.asClient()
            if user_c:
                strInfo += user_c.getIp()
            else:
                strInfo += "unknown"
            strInfo += "\n"
            strInfo += "Level: " + str(UserManager.getLevel(user)) + "\n"
            strInfo += field_function("DE", "Description")
            strInfo += field_function("SS", "Share size (bytes)")
            strInfo += field_function("SF", "Number of shared files")
            strInfo += field_function("AP", "Application")
            strInfo += field_function("VE", "Application version")
            strInfo += field_function("US", "Max upload speed (bytes/s)")
            strInfo += field_function("DS", "Max download speed (bytes/s)")
            strInfo += field_function("SL", "Max slots")
            strInfo += field_function("AS", "Speed limit for auto-slots (bytes/s)")
            strInfo += field_function("AM", "Minimum auto-slots")
            strInfo += field_function("EM", "E-mail")
            strInfo += field_function("HN", "Hubs where user is a normal user")
            strInfo += field_function("HR", "Hubs where user is registered")
            strInfo += field_function("HO", "Hubs where user is operator")
            strInfo += field_function("AW", "Away")
            strInfo += field_function("SU", "Protocol supports")

        else:
            # by IP
            users_ip = []
            entities = Hub.ClientMgr.getEntities()
            size = len(entities)
            if size > 0:
                for entity in entities:
                    user_c = entity.asClient()
                    if user_c and user_c.getIp() == params:
                        users_ip.append(entity)
            if len(users_ip) > 0:
                strInfo = "Users with the IP " + params + ":\n"
                for ent_w_ip in users_ip:
                    strInfo += ent_w_ip.getField("NI") + "\n"
            else:
                strInfo = "No user found with a nick, CID or IP matching " + params

        autil.hub_reply(ent, strInfo)
    else:
        hubtime = Hub.core.getUpTime()

        strInfo = "\n"
        strInfo += "Hub version: " + adchpp.cvar.versionString + "\n"
        strInfo += "Hub uptime: " + autil.format_seconds(hubtime) + "\n"

        strInfo += "\nADC and script commands: \n"
        sortedList = sorted(stats.items(),key=(lambda x: x[1]),reverse=True)
        for obj in sortedList:
            strInfo += str(obj[1]) + "\t" + obj[0] + "\n"
        sortedList = sorted(HubCommandStats.items(),key=(lambda x: x[1]),reverse=True)
        for obj in sortedList:
            strInfo += str(obj[1]) + "\t" + SettingManager.settings['cmd_chr'].value + obj[0] + "\n"

        strInfo += "\nDisconnect reasons: \n"
        for k in DCReasons:
            strInfo += str(DCReasons[k]) + "\t" + utilReasons[k] + "\n"
            
        strInfo += "\nSocket errors: \n"
            
        for k in socketErrors:
            strInfo += k + "\t" + str(socketErrors[k]) + "\n"        

        sockStats = Hub.SocketMgr.getStats()
            
        queued = Hub.ClientMgr.getQueuedBytes()
        queueBytes = sockStats.queueBytes
        queueCalls = sockStats.queueCalls
        sendBytes = sockStats.sendBytes
        sendCalls = sockStats.sendCalls
        recvBytes = sockStats.recvBytes
        recvCalls = sockStats.recvCalls

        strInfo += "\nBandwidth stats: \n"
        strInfo += str(adchpp.Util.formatBytes(queued)) + "\tBytes queued (" + str(adchpp.Util.formatBytes(queued / len(Hub.ClientMgr.getEntities()))) + "/user)\n"
        strInfo += str(adchpp.Util.formatBytes(queueBytes)) + "\tTotal bytes queued (" + str(adchpp.Util.formatBytes(queueBytes/hubtime)) + "/s)\n"
        strInfo += str(queueCalls) + "\tQueue calls (" + str(adchpp.Util.formatBytes(queueBytes/queueCalls)) + "/call)\n"
        strInfo += str(adchpp.Util.formatBytes(sendBytes)) + "\tTotal bytes sent (" + str(adchpp.Util.formatBytes(sendBytes/hubtime)) + "/s)\n"
        strInfo += str(sendCalls) + "\tSend calls (" + str(adchpp.Util.formatBytes(sendBytes/sendCalls)) + "/call)\n"
        strInfo += str(adchpp.Util.formatBytes(recvBytes)) + "\tTotal bytes received (" + str(adchpp.Util.formatBytes(recvBytes/hubtime)) + "/s)\n"
        strInfo += str(recvCalls) + "\tReceive calls (" + str(adchpp.Util.formatBytes(recvBytes/recvCalls)) + "/call)\n"

        strInfo += ScriptManager.getScriptsInfo()

        autil.hub_reply(ent, strInfo)

        dispatch_stats = True
        dispatch_cmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_HUB, ent.getSID())
        dispatch_cmd.addParam('+stats')
        ent.inject(dispatch_cmd)
        dispatch_stats = False

def onRecieve(ent, cmd, ok):
    addStats(cmd.getCommandString())

    if not ok:
        return ok

    c = ent.asClient()
    if not c:
        return False

        
    if cmd.getCommand() in command_contexts:
        allowed_type = command_contexts[cmd.getCommand()]
        if not re.match(allowed_type, cmd.getType()):
            autil.hub_reply(ent, "Invalid context for " + cmd.getCommandString())
            return False

    getCommand = cmd.getCommand()
    if c.getState() == adchpp.Entity.STATE_NORMAL and getCommand in restrCMDs:
        restricted = restrCMDs[getCommand]
        rest_level = 0
        if 'level' in restricted:
            if type(restricted['level']) is int:
                rest_level = restricted['level']
            elif type(restricted['level']) is SettingManager.Setting and type(restricted['level'].value) is int:
                rest_level = restricted['level'].value
        if UserManager.getLevel(ent) < rest_level:
            sendcmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_STA, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
            sendcmd.addParam(str(adchpp.AdcCommand.SEV_RECOVERABLE) + str(adchpp.AdcCommand.ERROR_COMMAND_ACCESS))
            if 'msg' in restricted:
                sendcmd.addParam("You are not allowed to " + restricted['msg'] + " in this hub")
            else:
                sendcmd.addParam("You are not allowed to use the command: " + str(cmd.getType()) + " in this hub")
            sendcmd.addParam("FC" + cmd.getFourCC())
            c.send(sendcmd)
            return False
        
    return True

def addStats(stat):
    global stats
    if stat in stats:
        stats[stat] += 1
    else:
        stats[stat] = 1

def addHubCmdStats(stat):
    if stat in HubCommandStats:
            HubCommandStats[stat] += 1
    else:
            HubCommandStats[stat] = 1