import pyadchpp as adchpp
import Hub
import HubBot
import json, os, math, threading

threadData = threading.local()

def receiveCommand(ent, cmd, ok, filter, callback):
    if not ok:
        return ok
    
    if filter != None and cmd.getCommand() != filter:
        return ok
    
    result = callback(ent, cmd, ok)

    if not result:
        cmd.setPriority(adchpp.AdcCommand.PRIORITY_IGNORE)
    
    return result
    
def handleCommand(callback,filter=None):
        return Hub.ClientMgr.signalReceive().connect(lambda ent, cmd, ok: receiveCommand(ent, cmd, ok, filter, callback))

def receiveState(ent, oldState, filter, callback):
    if ent.getState() != filter:
        return True
    
    return callback(ent, oldState)

def handleState(callback,filter=None):
    if filter:
        return Hub.ClientMgr.signalState().connect(lambda ent, oldState: receiveState(ent, oldState, filter, callback))
    else:
        return Hub.ClientMgr.signalState().connect(callback)

def addSupport(ext):
    Hub.ClientMgr.getEntity(adchpp.AdcCommand.HUB_SID).addSupports(adchpp.AdcCommand.toFourCC(ext))

def dump(client, code, reason, msg):

    cmdDump = adchpp.AdcCommand(adchpp.AdcCommand.CMD_QUI, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
    cmdDump.addParam(adchpp.AdcCommand.fromSID(client.getSID()))
    cmdDump.addParam("DI1")

    #cmdTest = adchpp.AdcCommand(adchpp.AdcCommand.CMD_QUI, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
    #cmdTest.addParam(adchpp.AdcCommand.fromSID(client.getSID()))
    
    if callable(msg):
        msg(cmdDump)
        #msg(cmdTest)
        msg = cmdDump.getParam("MS",0)[1]
    else:
        #cmdTest.addParam("MS", str(msg))
        cmdDump.addParam("MS", str(msg))

    #Hub.ClientMgr.sendToAll(cmdTest.getBuffer())

    if client.getState() == adchpp.Entity.STATE_NORMAL:
        cmdSend = adchpp.AdcCommand(adchpp.AdcCommand.CMD_STA, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
        cmdSend.addParam(str(adchpp.AdcCommand.SEV_FATAL) + str(code))
        cmdSend.addParam(str(msg))
        client.send(cmdSend)

    client.send(cmdDump)
    client.disconnect(reason)
    
    """answer = adchpp.AdcCommand(adchpp.AdcCommand.CMD_STA, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)

    answer.addParam(str(adchpp.AdcCommand.SEV_FATAL) + str(code)).addParam(str(msg))
        
    client.send(answer)
    client.disconnect(0)"""
    return False

def info(msg):
    cmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_INFO, adchpp.AdcCommand.HUB_SID)
    cmd.addParam(str(msg))
    return cmd

def pm_direct(msg, senderEnt, sender, recipient):
    command = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_DIRECT, sender)
    command.addParam(str(msg))
    command.addParam("PM", adchpp.AdcCommand.fromSID(senderEnt))
    if recipient:
        command.setTo(recipient)
    return command

def pm_echo(msg, senderEnt, sender, recipient):
    command = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_ECHO, sender)
    command.addParam(str(msg))
    command.addParam("PM", adchpp.AdcCommand.fromSID(senderEnt))
    if recipient:
        command.setTo(recipient)
    return command

def hub_reply(ent, msg):
    if hasattr(threadData,'replyFrom') and (threadData.replyFrom != adchpp.AdcCommand.INVALID_SID and threadData.replyFrom != adchpp.AdcCommand.HUB_SID):
        replyEnt = Hub.ClientMgr.getEntity(threadData.replyFrom)
        if replyEnt:
            ent.send( pm_direct(msg, replyEnt.SID, HubBot.bot.getSID(), ent.SID) ) #HUB_SID cannot reply in pm windows, so it has to be hub bot at least so users don't get confused
            return
            
    ent.send( info(msg) )

def hub_broadcast(msg):
    cmd = info(msg)
    Hub.ClientMgr.sendToAll( cmd.getBuffer() )

def hub_broadcast_ME(msg):
    cmd = info(msg)
    cmd.addParam("ME", "1")
    Hub.ClientMgr.sendToAll( cmd.getBuffer() )

def ent_reply(ent, msg):
    if hasattr(threadData,'replyFrom') and (threadData.replyFrom != adchpp.AdcCommand.INVALID_SID and threadData.replyFrom != adchpp.AdcCommand.HUB_SID):
        replyEnt = Hub.ClientMgr.getEntity(threadData.replyFrom)
        if replyEnt:
            ent.inject( pm_echo(msg, ent.SID, ent.SID, replyEnt.SID) )
            return
    cmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_BROADCAST, ent.getSID())
    cmd.addParam(str(msg))
    ent.inject( cmd )

def ent_reply_ME(ent, msg):
    if hasattr(threadData,'replyFrom') and (threadData.replyFrom != adchpp.AdcCommand.INVALID_SID and threadData.replyFrom != adchpp.AdcCommand.HUB_SID):
        replyEnt = Hub.ClientMgr.getEntity(threadData.replyFrom)
        if replyEnt:
            ent.inject( pm_echo(msg, ent.SID, ent.SID, replyEnt.SID) )
            return
    cmd = adchpp.AdcCommand(adchpp.AdcCommand.CMD_MSG, adchpp.AdcCommand.TYPE_BROADCAST, ent.getSID())
    cmd.addParam(str(msg))
    cmd.addParam("ME","1")
    ent.inject( cmd )

def toUTF8(x):
    if type(x) is unicode:
        return(x.encode('utf-8',errors='replace'))
    else:
        return(unicode(x, 'utf-8', errors="replace"))

def log(area, msg):
    print( "[" + str(area) + "]: " + str(msg))
    Hub.LogMgr.log(str(area), str(msg))

def fallbackdict(dict):
    def __init__(self, fallback):
        self.fallback = fallback
        
    def __getitem__(self, key):
        try:
            return super(fallbackdict, self).__getitem__(self, key)
        except KeyError:
            return self.fallback.__getitem__(self, key)

def toInt(x):
    try:
        return int(x)
    except:
        return 0

def toNumber(x):
    try:
        return float(x)
    except:
        return 0

def format_seconds(t):
    t_d = math.floor(t / (60*60*24))
    t_h = math.floor(t / (60*60)) % 24
    t_m = math.floor(t / 60) % 60
    t_s = t % 60

    return "%d days, %d hours, %d minutes and %d seconds" % (t_d, t_h, t_m, t_s)

def format_minutes(t):
    t_d = math.floor(t / (60*60*24))
    t_h = math.floor(t / (60*60)) % 24
    t_m = math.floor(t / 60) % 60

    return "%d days, %d hours and %d minutes" % (t_d, t_h, t_m)

def parseJsonFile(fileName):
    if os.path.exists(fileName):
        try:
            loaded = json.load( open(fileName, 'r') )
            #log(__name__, fileName + ' file read and decodeded successfully')
            return(loaded)
        except Exception as err:
            log(__name__, fileName + ' file could not be read (' + str(err) + ')')
    else:
        log(__name__, fileName + ' could not be found for reading')

def saveJsonFile(data, fileName):
    dir = os.path.dirname(fileName)
    if not os.path.exists(dir):
        os.makedirs(dir)

    try:
        file = open(fileName, 'w')
        if not file:
            raise Exception("Could not open " + filename + " for writing.")
        json.dump(data, file)
        #log(__name__, fileName + ' file saved')
        return True
    except Exception as err:
        log(__name__, 'could not save ' + fileName + ' file: ' + str(err))
        return False
