import CommandManager
import pyadchpp as adchpp
import Hub
import autil
from ScriptManager import ScriptBase
import codecs

import re, sys

dataDir = "PyScripts/data/LaTeX/"

class Script(ScriptBase):
    superscripts = {}
    subscripts = {}
    textbb = {}
    textbf = {}
    textit = {}
    textcal = {}
    textfrak = {}
    textmono = {}
    latex_symbols = []

    #utf-8 encoded chars
    hat = ['\xcc\x82']
    widehat = ['\xef\xb8\xa0','\xef\xb8\xa1']
    check = ['\xcc\x8c']
    tilde = ['\xcc\x83']
    widetilde = ['\xef\xb8\xa2','\xef\xb8\xa3']
    acute = ['\xcc\x81']
    grave = ['\xcc\x80']
    dot = ['\xcc\x87']
    ddot = ['\xcc\x88']
    breve = ['\xcc\x86']
    bar = ['\xcc\x84']
    vec = ['\xe2\x83\x97']

    def __init__(self):
        self.load_data()

        self.latexCmd = CommandManager.Command(key='LaTeX', alias=['TeX'],category="ScriptCommands",command=self.convertToLatex, helpTxt="string - converts LaTeX formatted string to proper unicode", helpTxtLong=self.help)

        CommandManager.add(self.latexCmd)

    def unload(self):
        self.latexCmd.release()

    def reload(self):
        self.latexCmd.command = self.convertToLatex

    def help(self):
        strHelp = "List of available modifiers:\n\n"
        strHelp += "\\hat{x} will return " + self.apply_combiner('\\hat{x}', "\\hat", self.hat) + "\n"
        strHelp += "\\widehat{xx} will return " + self.apply_combiner('\\widehat{xx}', "\\widehat", self.widehat) + "\n"
        strHelp += "\\check{x} will return " + self.apply_combiner('\\check{x}', "\\check", self.check) + "\n"
        strHelp += "\\tilde{x} will return " + self.apply_combiner('\\tilde{x}', "\\tilde", self.tilde) + "\n"
        strHelp += "\\widetilde{xx} will return " + self.apply_combiner('\\widetilde{xx}', "\\widetilde", self.widetilde) + "\n"
        strHelp += "\\grave{x} will return " + self.apply_combiner('\\grave{x}', "\\grave", self.grave) + "\n"
        strHelp += "\\dot{x} will return " + self.apply_combiner('\\dot{x}', "\\dot", self.dot) + "\n"
        strHelp += "\\ddot{x} will return " + self.apply_combiner('\\ddot{x}', "\\ddot", self.ddot) + "\n"
        strHelp += "\\breve{x} will return " + self.apply_combiner('\\breve{x}', "\\breve", self.breve) + "\n"
        strHelp += "\\bar{x} will return " + self.apply_combiner('\\bar{x}', "\\bar", self.bar) + "\n"
        strHelp += "\\vec{x} will return " + self.apply_combiner('\\vec{x}', "\\vec", self.vec) + "\n\n"

        strHelp += "x^x or x^{xx} will superscript some characters " + self.apply_modifier("x^x or x^{xx}", "^", self.superscripts) + "\n"
        strHelp += "x_x or x_{xx} will subscript some characters " + self.apply_modifier("x_x or x_{xx}", "_", self.subscripts) + "\n\n"

        strHelp += "\\bb{x} will textbord bold some characters:\n"
        for x in self.textbb:
            strHelp += '\t' + x + "  -  " + self.textbb[x] + "\n"

        strHelp += '\n'

        strHelp += "\\bb{x} will textbord bold some characters:\n"
        for x in self.textbb:
            strHelp += '\t' + x + "  -  " + self.textbb[x] + "\n"

        strHelp += '\n'

        strHelp += "The following is a list of symbols available:\n"
        for x in self.latex_symbols:
            strHelp += '\t' + x[0] + "  -  " +x[1] + "\n"

        strHelp += '\n'

        return strHelp


    def convertToLatex(self, ent, cmd, params, public):
        if len(params) == 0:
            autil.hub_reply(ent, "You need to supply a string to convert")
            return

        params = params[0]

        params = self.convert(params)

        if public:
            autil.ent_reply(ent, params)
        else:
            autil.hub_reply(ent, params)


    def convert(self, s):
        s = self.convert_latex_symbols(s)
        s = self.process_starting_modifiers(s)
        s = self.apply_all_modifiers(s)
        return s

    # Replace each "\alpha", "\beta" and similar latex symbols with
    # their unicode representation.
    def convert_latex_symbols(self, s):
        for (code, val) in self.latex_symbols:
            s = s.replace(code, val)
        return s

    # If s start with "it ", "cal ", etc. then make the whole string
    # italic, calligraphic, etc.
    def process_starting_modifiers(self, s):
        s = re.sub("^bb ", r"\\bb{", s) #some not not supported in utf-8
        #s = re.sub("^bf ", r"\\bf{", s) #not supported in utf-8
        #s = re.sub("^it ", r"\\it{", s) #not supported in utf-8
        #s = re.sub("^cal ", r"\\cal{", s) #not supported in utf-8
        s = re.sub("^frak ", r"\\frak{", s) #some not not supported in utf-8
        #s = re.sub("^mono ", r"\\mono{", s) #not supported in utf-8
        return s

    def apply_all_modifiers(self, s):
        s = self.apply_modifier(s, "^", self.superscripts)
        s = self.apply_modifier(s, "_", self.subscripts)
        s = self.apply_modifier(s, "\\bb", self.textbb)
        #s = self.apply_modifier(s, "\\bf", self.textbf)
        #s = self.apply_modifier(s, "\\it", self.textit)
        #s = self.apply_modifier(s, "\\cal", self.textcal)
        s = self.apply_modifier(s, "\\frak", self.textfrak)
        #s = self.apply_modifier(s, "\\mono", self.textmono)

        s = self.apply_combiner(s, "\\hat", self.hat)
        s = self.apply_combiner(s, "\\widehat", self.widehat)
        s = self.apply_combiner(s, "\\check", self.check)
        s = self.apply_combiner(s, "\\tilde", self.tilde)
        s = self.apply_combiner(s, "\\widetilde", self.widetilde)
        s = self.apply_combiner(s, "\\acute", self.acute)
        s = self.apply_combiner(s, "\\grave", self.grave)
        s = self.apply_combiner(s, "\\dot", self.dot)
        s = self.apply_combiner(s, "\\ddot", self.ddot)
        s = self.apply_combiner(s, "\\breve", self.breve)
        s = self.apply_combiner(s, "\\bar", self.bar)
        s = self.apply_combiner(s, "\\vec", self.vec)
        return s

    def apply_combiner(self, text, modifier, chars):
        text = text.replace(modifier, "^")
        newtext = ""

        mode_normal, mode_modified, mode_long = range(3)
        mode = mode_normal

        counter = 0
        for ch in text:
            nextchar = chars[counter]

            if mode == mode_normal and ch == '^':
                mode = mode_modified
                continue
            elif mode == mode_modified and ch == '{':
                mode = mode_long
                continue
            elif mode == mode_modified:
                newtext += ch + nextchar
                counter = (counter+1)%len(chars)
                mode = mode_normal
                continue
            elif mode == mode_long and ch == '}':
                mode = mode_normal
                continue
            
            if mode == mode_normal:
                newtext += ch
            else:
                newtext += ch + nextchar
                counter = (counter+1)%len(chars)

        return newtext

    # Example: modifier = "^", D = self.superscripts
    # This will search for the ^ signs and replace the next
    # digit or (digits when {} is used) with its/their uppercase representation.
    def apply_modifier(self, text, modifier, D):
        text = text.replace(modifier, "^")
        newtext = ""
        mode_normal, mode_modified, mode_long = range(3)
        mode = mode_normal
        for ch in text:
            if mode == mode_normal and ch == '^':
                mode = mode_modified
                continue
            elif mode == mode_modified and ch == '{':
                mode = mode_long
                continue
            elif mode == mode_modified:
                newtext += D.get(ch, ch)
                mode = mode_normal
                continue
            elif mode == mode_long and ch == '}':
                mode = mode_normal
                continue

            if mode == mode_normal:
                newtext += ch
            else:
                newtext += D.get(ch, ch)
        return newtext

    def load_data(self):
        self.load_symbols()
        self.load_dict(dataDir + "subscripts", self.subscripts)
        self.load_dict(dataDir + "superscripts", self.superscripts)
        self.load_dict(dataDir + "textbb", self.textbb)
        #self.load_dict(dataDir + "textbf", self.textbf)
        #self.load_dict(dataDir + "textit", self.textit)
        #self.load_dict(dataDir + "textcal", self.textcal)
        self.load_dict(dataDir + "textfrak", self.textfrak)
        #self.load_dict(dataDir + "textmono", self.textmono)

    def load_dict(self, filename, D):
        with open(filename, "r") as f:
            line = f.readline()
            while line != "":
                words = line.split()
                code = words[0]
                val = words[1]
                D[code] = val
                line = f.readline()

    def load_symbols(self):
        with open(dataDir + "symbols", "r") as f:
            line = f.readline()
            while line != "":
                words = line.split()
                code = words[0]
                val = words[1]
                self.latex_symbols.append((code, val))
                line = f.readline()


    
