import SettingManager
import CommandManager
import UserManager
import HubBot
import pyadchpp as adchpp
import general
import Hub
from ScriptManager import ScriptBase
import autil


class Script(ScriptBase):
    def __init__(self):
        self.bot = None

        self.botCID = SettingManager.Setting(key="opchat_cid", alias=['opchatcid'], helpTxt="CID of the bot, restart the opchat after the change",value=adchpp.CID.generate().toBase32(), validate=HubBot.validateCID)
        self.botName = SettingManager.Setting(key="opchat_name", alias=['opchatnick','opchatni'], change=self.botNameChange, helpTxt="name of the opchat bot",value="- OPChat -", validate=general.validate_ni)
        self.botDesc = SettingManager.Setting(key="opchat_description", alias=['opchatdescr','opchatde'], change=self.botDescChange, helpTxt="description of the opchat bot",value="", validate=general.validate_de)
        self.botEmail = SettingManager.Setting(key="opchat_email", alias=['opchatmail','opchatem'], change=self.botEmailChange, helpTxt="e-mail of the opchat bot",value="", validate=general.validate_de)

        SettingManager.add(self.botCID)
        SettingManager.add(self.botName)
        SettingManager.add(self.botDesc)
        SettingManager.add(self.botEmail)

        #self.bot = HubBot.makeBot(cid=self.botCID.value,name=self.botName.value,desc=self.botDesc.value,email=self.botEmail.value)
        self.bot = Hub.ClientMgr.createBot(lambda bot, buffer: self.onMSG(None, adchpp.AdcCommand(buffer), True))
        self.bot.setCID(adchpp.CID(self.botCID.value))
        self.bot.setField("ID", self.botCID.value)
        self.bot.setField("NI", self.botName.value)
        self.bot.setField("DE", self.botDesc.value)
        self.bot.setField("EM", self.botEmail.value)
        self.bot.setFlag(adchpp.Entity.FLAG_REGISTERED)
        self.bot.setFlag(adchpp.Entity.FLAG_OP)
        self.bot.setFlag(adchpp.Entity.FLAG_SU)
        self.bot.setFlag(adchpp.Entity.FLAG_OWNER)

        Hub.ClientMgr.regBot(self.bot)

        self.msgCMD = autil.handleCommand(self.onMSG, adchpp.AdcCommand.CMD_MSG)

    def unload(self):
        del(self.msgCMD)

        self.botCID.release()
        self.botName.release()
        self.botDesc.release()
        self.botEmail.release()

        self.bot.disconnect(adchpp.Util.REASON_PLUGIN)

    def botNameChange(self, oldValue, newValue):
        if self.bot:
            self.bot.setField("NI", newValue)
            infCMD = adchpp.AdcCommand(adchpp.AdcCommand.CMD_INF, adchpp.AdcCommand.TYPE_BROADCAST, self.bot.getSID())
            infCMD.addParam("NI", newValue)
            Hub.ClientMgr.sendToAll(infCMD.getBuffer())

    def botDescChange(self, oldValue, newValue):
        if self.bot:
            self.bot.setField("DE", newValue)
            infCMD = adchpp.AdcCommand(adchpp.AdcCommand.CMD_INF, adchpp.AdcCommand.TYPE_BROADCAST, self.bot.getSID())
            infCMD.addParam("DE", snewValue)
            Hub.ClientMgr.sendToAll(infCMD.getBuffer())

    def botEmailChange(self, oldValue, newValue):
        if self.bot:
            self.bot.setField("EM", newValue)
            infCMD = adchpp.AdcCommand(adchpp.AdcCommand.CMD_INF, adchpp.AdcCommand.TYPE_BROADCAST, self.bot.getSID())
            infCMD.addParam("EM", newValue)
            Hub.ClientMgr.sendToAll(infCMD.getBuffer())

    def onMSG(self, ent, cmd, ok):
        if not ent:
            return True

        if cmd.getCommand() == adchpp.AdcCommand.CMD_MSG:
            if cmd.getTo() == self.bot.getSID(): 

                if UserManager.getLevel(ent) < SettingManager.settings['level_op'].value:
                    ent.send( autil.pm_direct("You don't have sufficient rights to access this chat", self.bot.getSID(), ent.getSID()) )
                    return False

                result = CommandManager.onMSG(ent, cmd, ok)

                if result:
                    cmd.setFrom(ent.getSID())
                    cmd.delParam('PM',0)
                    cmd.addParam('PM'+adchpp.AdcCommand.fromSID(self.bot.getSID()))
                    entities = Hub.ClientMgr.getEntities()

                    for send_ent in entities:
                        if send_ent.SID == ent.SID or send_ent.SID == self.bot.SID or send_ent.SID == adchpp.AdcCommand.HUB_SID:
                            continue
                        if UserManager.getLevel(send_ent) >= SettingManager.settings['level_op'].value:
                            cmd.setTo(send_ent.getSID())
                            send_ent.send(cmd)

                cmd.setTo(self.bot.getSID())

                return True

        return True